/*=======================================================================
   Name:         Flexbar.cpp
   Authors:      Matthias Dodt and Johannes Roehr

   Description:  Flexbar - flexible barcode and adapter removal
   Version:      2.32
   Copyright:    GPL version 3

   SeqAn lib:    post release 1.3.1, revision 13058 on October 9, 2012
   TBB   lib:    version 4.0 update 5, stable release June 13, 2012
========================================================================*/


#include "Flexbar.h"
#include "Options.h"

#include "Enums.h"
#include "MultiplexedInputFilter.h"
#include "MultiplexedOutputFilter.h"
#include "MultiplexedAlignmentFilter.h"
#include "AlignmentAlgorithm.h"


int main(int argc, const char* argv[]){
	
	using namespace std;
	using namespace flexbar;
	
	using seqan::ArgumentParser;
	using seqan::CharString;
	
	
	// performTest();
	
	const string version = "2.32";
	const string date    = "February 17, 2013";
	
	ArgumentParser parser("flexbar");
	defineOptionsAndHelp(parser, version, date);
	
	Options o;
	
	parseCommandLine(o, parser, version, argc, argv);
	
	loadProgramOptions(o, parser);
	loadBarcodeAndAdapters(o);
	
	
	time_t t_start;
	time(&t_start);
	
	tbb::task_scheduler_init init_serial(o.nThreads);
	tbb::pipeline pipe;
	
	MultiplexedInputFilter<CharString, CharString > input_filter(o);
	
	cout << "\nProcessing reads ..." << flush;
	
	if(o.logLevel != NONE) cout << "\n\nGenerating " << o.log_level << " verbose output.\n" << endl;
	
	
	MultiplexedAlignmentFilter<CharString, CharString, AlignmentAlgorithm<CharString> > filter(o, &o.barcodes, &o.adapters);
	MultiplexedOutputFilter<CharString, CharString> output_filter(o, &o.barcodes, &o.adapters);
	
	pipe.add_filter(input_filter);
	pipe.add_filter(filter);
	pipe.add_filter(output_filter);
	
	pipe.run(o.nThreads);
	
	if(o.logLevel == TAB) cout << "\n";
	cout << "done.\n" << endl;
	
	printComputationTime(t_start);
	
	
	// barcode and adapter removal statistics
	
	if(o.writeLengthDist) output_filter.writeLengthDist();
	
	if(o.adapRm != AOFF){
		output_filter.printAdapterRemovalStats();
		filter.printAdapterOverlapStats();
	}
	
	output_filter.printFileSummary();
	
	
	const unsigned long nReads     = input_filter.getNrProcessedReads();
	const unsigned long nGoodReads = output_filter.getNrGoodReads();
	const unsigned long uncalled   = input_filter.getNrUncalledReads();
	const unsigned long uPairs     = input_filter.getNrUncalledPairedReads();
	
	cout << "Filtering statistics\n";
	cout << "====================\n";
	cout << "Processed reads                   " << nReads << endl;
	cout << "  skipped due to uncalled bases   ";
	
	if(o.runType == PAIRED || o.runType == PAIRED_BARCODED){
		cout << 2 * uPairs;
		
		if(uncalled > 0)
		cout << "   (" << uncalled << " uncalled in " << uPairs << " pairs)";
		cout << endl;
	}
	else cout << uncalled << endl;
	
	if(o.phred_preQual > 0)
	cout << "  trimmed due to low quality      " << input_filter.getNrLowPhredReads() << endl;
	
	if(o.adapRm != AOFF)
	cout << "  short prior adapter removal     " << filter.getNrPreShortReads() << endl;
	
	cout << "  finally skipped short reads     " << output_filter.getNrShortReads() << endl;
	cout << "Discarded reads overall           " << (nReads - nGoodReads) << endl;
	
	if(nReads > 0){
		cout << "Remaining reads                   " << nGoodReads << "   (" << fixed <<
		        setprecision(2) << 100 * nGoodReads / nReads << "% of input reads)\n" << endl;
	}
	
	printCompletedMessage(o);
	
	
	return 0;
}

